# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from importlib import metadata

project = "quattro"
copyright = "2025, Tin Tvrtkovic"
author = "Tin Tvrtkovic"

# The full version, including alpha/beta/rc tags.
release = metadata.version("quattro")
# The short X.Y version.
version = release.rsplit(".", 1)[0]

if "dev" in release:
    release = version = "UNRELEASED"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc", "myst_parser", "sphinx.ext.napoleon"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_theme_options = {
    "light_css_variables": {
        "font-stack": "Inter,sans-serif",
        "font-stack--monospace": "'Ubuntu Mono', monospace",
        "code-font-size": "90%",
        "color-highlight-on-target": "transparent",
    },
    "dark_css_variables": {"color-highlight-on-target": "transparent"},
}

myst_heading_anchors = 3
myst_enable_extensions = ["attrs_block"]
autodoc_typehints = "description"
autoclass_content = "both"



from asyncio import (
    CancelledError,
    TimeoutError,
    create_task,
    current_task,
    get_running_loop,
    sleep,
)
from time import monotonic, time

import pytest

from quattro import CancelScope, fail_after
from quattro._cancelscope import _is_311_or_later


async def test_fail_after():
    start = time()
    t = 0.5
    with pytest.raises(TimeoutError), fail_after(t) as cancel_scope:
        await sleep(1.0)

    assert cancel_scope.cancelled_caught
    spent = time() - start
    assert t <= spent <= t + 0.01


async def test_fail_after_not_triggered():
    start = time()
    with fail_after(0.2) as cancel_scope:
        await sleep(0.1)

    spent = time() - start

    assert not cancel_scope.cancelled_caught
    assert 0.1 <= spent <= 0.15

    await sleep(0.3)  # Sleep to possibly fail due to uncancelled timers


async def test_fail_after_not_triggered_no_await():
    with fail_after(0.2) as cancel_scope:
        pass
    assert not cancel_scope.cancelled_caught
    await sleep(0.3)


async def test_fail_after_externally_cancelled() -> None:
    cancel_scope = None

    async def task():
        nonlocal cancel_scope
        with fail_after(0.2) as cancel_scope:
            # Will be externally cancelled here.
            await sleep(1)

    t = create_task(task())
    await sleep(0.1)
    assert not t.done()
    t.cancel()
    with pytest.raises(CancelledError):
        await t

    assert cancel_scope is not None
    assert not cancel_scope.cancelled_caught


async def test_fail_after_inner_exception():
    cancel_scope = None

    async def task():
        nonlocal cancel_scope
        with fail_after(0.3) as cancel_scope:
            await sleep(0.2)
            1 / 0  # noqa: B018

    t = create_task(task())
    await sleep(0.1)
    assert not t.done()
    with pytest.raises(ZeroDivisionError):
        await t

    assert cancel_scope is not None
    assert not cancel_scope.cancelled_caught

    await sleep(0.3)  # Catch lingering timers


async def test_fail_after_nested_outer_shorter():
    """Nested fail_after blocks work properly."""
    checkpt_1 = 0
    checkpt_2 = 0
    start = time()
    with pytest.raises(TimeoutError), fail_after(0.1) as outer:
        checkpt_1 = 1
        with fail_after(0.2) as inner:
            checkpt_2 = 1
            await sleep(0.3)
            pytest.fail("Not cancelled")
        pytest.fail("Not cancelled")

    spent = time() - start
    assert 0.1 <= spent <= 0.15
    assert checkpt_1 == 1
    assert checkpt_2 == 1
    assert outer.cancelled_caught
    assert not inner.cancelled_caught


async def test_fail_after_nested_inner_shorter():
    """Nested fail_after blocks work properly."""
    checkpt_1 = 0
    checkpt_2 = 0
    checkpt_3 = 0
    start = time()
    with pytest.raises(TimeoutError), fail_after(0.2) as outer:
        checkpt_1 = 1
        try:
            with fail_after(0.1) as inner:
                checkpt_2 = 1
                await sleep(0.3)
                pytest.fail("Not cancelled")
        except TimeoutError:
            checkpt_3 = 1
            await sleep(0.2)
        else:
            pytest.fail("No timeout")

    spent = time() - start

    assert 0.2 <= spent <= 0.25
    assert checkpt_1 == 1
    assert checkpt_2 == 1
    assert checkpt_3 == 1
    assert outer.cancelled_caught
    assert inner.cancelled_caught


async def test_fail_after_nested_happy() -> None:
    """Nested fail_after blocks work properly."""
    checkpt_1 = 0
    checkpt_2 = 0
    checkpt_3 = 0
    start = monotonic()
    with fail_after(0.2) as outer:
        checkpt_1 = 1
        with fail_after(0.1) as inner:
            checkpt_2 = 1
        checkpt_3 = 1

    spent = monotonic() - start
    assert 0 <= spent <= 0.005  # Add a little buffer for PyPy
    assert checkpt_1 == 1
    assert checkpt_2 == 1
    assert checkpt_3 == 1
    assert not outer.cancelled_caught
    assert not inner.cancelled_caught


async def test_fail_after_cancel_myself():
    async def task():
        with fail_after(0.5) as cancel_scope:
            await sleep(0.1)
            cancel_scope.cancel()

    t = create_task(task())
    await sleep(0.05)
    assert not t.done()
    await sleep(0.1)
    assert t.done()

    with pytest.raises(CancelledError):
        await t

    await sleep(0.5)


async def test_fail_after_precancel() -> None:
    """Precancelling works."""
    cancel_scope = fail_after(0.5)

    async def task():
        with cancel_scope:
            await sleep(0.1)

    cancel_scope.cancel()

    t = create_task(task())
    await sleep(0.01)
    assert t.done()

    assert isinstance(t.exception(), TimeoutError)

    await sleep(0.5)


async def test_fail_after_move_deadline():
    start = time()
    with pytest.raises(TimeoutError), fail_after(0.1) as cancel_scope:
        await sleep(0.01)
        cancel_scope.deadline += 0.1
        await sleep(1)

    spent = time() - start

    assert 0.2 <= spent <= 0.21


async def test_fail_after_remove_deadline():
    start = time()
    with fail_after(0.2) as cancel_scope:
        await sleep(0.1)
        cancel_scope.deadline = None
        await sleep(0.3)

    spent = time() - start
    assert 0.4 <= spent <= 0.41


async def test_fail_after_move_deadline_to_past():
    start = time()
    with pytest.raises(TimeoutError), fail_after(0.2) as cancel_scope:
        await sleep(0.1)
        cancel_scope.deadline -= 0.1
        await sleep(0.3)

    spent = time() - start
    assert 0.1 <= spent <= 0.15


async def test_fail_after_move_noop():
    """Nothing bad happens if the deadline is moved by 0."""
    start = time()
    with pytest.raises(TimeoutError), fail_after(0.2) as cancel_scope:
        await sleep(0.1)
        cancel_scope.deadline += 0.0
        await sleep(0.3)

    spent = time() - start
    assert 0.2 <= spent <= 0.205

    await sleep(0.1)  # Any lingering errors


@pytest.mark.skipif(not _is_311_or_later, reason="3.11+ only")
async def test_fail_after_inner_unrelated_exc() -> None:
    """Nested scopes handle current task cancellation properly."""
    with pytest.raises(TimeoutError), CancelScope() as first:
        first._raise_on_cancel = True
        with fail_after(0.7) as second:
            try:
                with fail_after(1) as third:
                    second.cancel()
                    first.cancel()
                    raise ValueError()
            except ValueError:
                assert not third.cancelled_caught
                await sleep(0.3)
    assert not second.cancelled_caught
    assert first.cancelled_caught
    curr = current_task()
    assert curr and curr.cancelling() == 0


async def test_deadline_expired_on_enter() -> None:
    """Past deadlines work on entering the scope."""
    with pytest.raises(TimeoutError), fail_after(0):
        await sleep(1)


@pytest.mark.skipif(not _is_311_or_later, reason="3.11+ only")
async def test_double_cancel_protection():
    """Test that calling cancel() multiple times only cancels the task once."""

    async def mock_task():
        with fail_after(1000.0) as scope:
            # Call cancel twice in succession
            scope.cancel()
            scope.cancel()

            # Check that task is only cancelled once
            curr = current_task()
            assert curr and curr.cancelling() == 1

            await sleep(0.1)

    with pytest.raises(TimeoutError):
        await mock_task()


async def test_deadline_outside_context():
    """Setting the deadline before entering the context manager does nothing."""

    scope = CancelScope()
    scope._raise_on_cancel = True
    scope.deadline = get_running_loop().time() + 0.1
    assert scope._timeout_handler is None

    with pytest.raises(TimeoutError), scope:
        await sleep(0.2)


async def test_set_deadline_inside() -> None:
    with pytest.raises(TimeoutError), CancelScope(None) as scope:
        scope._raise_on_cancel = True
        scope.deadline = get_running_loop().time() - 1
        await sleep(0)



from asyncio import TimeoutError, get_running_loop, sleep

import pytest

from quattro import (
    TaskGroup,
    fail_at,
    get_current_effective_deadline,
    move_on_at,
)


async def test_effective_deadline():
    assert get_current_effective_deadline() == float("inf")

    deadline = get_running_loop().time() + 10
    with move_on_at(deadline):
        assert get_current_effective_deadline() == deadline

    assert get_current_effective_deadline() == float("inf")


async def test_nested_deadlines():
    assert get_current_effective_deadline() == float("inf")

    deadline = get_running_loop().time() + 10
    with move_on_at(deadline):
        assert get_current_effective_deadline() == deadline

        fail_deadline = deadline - 5
        with fail_at(fail_deadline):
            assert get_current_effective_deadline() == fail_deadline

        assert get_current_effective_deadline() == deadline

        future_deadline = deadline + 5
        with move_on_at(future_deadline):
            assert get_current_effective_deadline() == deadline

    assert get_current_effective_deadline() == float("inf")


async def test_nested_deadlines_error():
    """Nesting move_on and fail_at produces correct deadlines."""
    assert get_current_effective_deadline() == float("inf")

    deadline = get_running_loop().time() + 1
    with move_on_at(deadline):
        assert get_current_effective_deadline() == deadline

        fail_deadline = deadline - 0.7
        with pytest.raises(TimeoutError), fail_at(fail_deadline):
            assert get_current_effective_deadline() == fail_deadline
            await sleep(1.0)

        assert get_current_effective_deadline() == deadline

        future_deadline = deadline + 1
        with move_on_at(future_deadline):
            assert get_current_effective_deadline() == deadline

    assert get_current_effective_deadline() == float("inf")


async def test_moved_nested_deadlines():
    assert get_current_effective_deadline() == float("inf")

    deadline = get_running_loop().time() + 10

    with move_on_at(deadline) as cancel_scope:
        assert get_current_effective_deadline() == deadline

        new_deadline = deadline + 1
        cancel_scope.deadline = new_deadline
        assert get_current_effective_deadline() == new_deadline

        fail_deadline = deadline - 5
        with fail_at(fail_deadline):
            assert get_current_effective_deadline() == fail_deadline

        assert get_current_effective_deadline() == new_deadline

        future_deadline = deadline + 5
        with move_on_at(future_deadline):
            assert get_current_effective_deadline() == new_deadline

        cancel_scope.deadline = None
        assert get_current_effective_deadline() == float("inf")

    assert get_current_effective_deadline() == float("inf")


async def test_taskgroup_deadlines():
    assert get_current_effective_deadline() == float("inf")

    async with TaskGroup() as tg:

        async def task():
            assert get_current_effective_deadline() == float("inf")

        await tg.create_task(task())

    deadline = get_running_loop().time() + 1

    with move_on_at(deadline):
        assert get_current_effective_deadline() == deadline
        async with TaskGroup() as tg:

            async def task():
                assert get_current_effective_deadline() == deadline

                new_deadline = deadline - 0.1
                with move_on_at(new_deadline):
                    assert get_current_effective_deadline() == new_deadline

                assert get_current_effective_deadline() == deadline

            await tg.create_task(task())


async def test_precancelled_deadline() -> None:
    """If a cancel scope is precancelled, the deadline is correct."""
    assert get_current_effective_deadline() == float("inf")

    deadline = get_running_loop().time() + 10
    scope = move_on_at(deadline)
    scope.cancel()
    assert get_current_effective_deadline() == float("inf")
    with scope:
        assert get_current_effective_deadline() <= get_running_loop().time()
    assert get_current_effective_deadline() == float("inf")
    # assert get_current_effective_deadline() == deadline

    assert get_current_effective_deadline() == float("inf")



import sys
from collections.abc import Awaitable, Callable
from contextlib import AbstractAsyncContextManager, AsyncExitStack
from contextvars import ContextVar
from typing import Final, TypeVar, Union, overload

if sys.version_info < (3, 10):
    from typing_extensions import Concatenate, ParamSpec
else:
    from typing import Concatenate, ParamSpec

P = ParamSpec("P")
T = TypeVar("T")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
Aw = TypeVar("Aw", bound=Awaitable)


class Deferrer(AsyncExitStack):
    """A decorator and class to enable deferring functions until the end of a coroutine.

    First, apply the Deferrer.enable decorator to a coroutine function.
    The coroutine function will receive an instance of `Deferrer` as its first
    positional argument.

    This instance of `Deferrer` can be used to enter async context managers; these
    context managers will be exited after the coroutine finishes.

    Example:
        >>> from quattro import Deferrer
        >>> @Deferrer.enable
        ... async def test(defer: Deferrer) -> None:
        ...     tg = defer(TaskGroup())  # This TaskGroup will be exited after return
        ...

    `Deferrer` is a subclass of `contextlib.AsyncExitStack`, so it supports its full
    API:

    * `AsyncExitStack.enter_async_context` (equivalent to just `Defer.__call__`
      shown above)
    * `AsyncExitStack.push_async_callback`
    * `AsyncExitStack.push_async_exit`
    * `BaseExitStack.enter_context`
    * `BaseExitStack.callback`
    * `BaseExitStack.push`

    Loosely inspired by the Go `defer` statement. (https://go.dev/tour/flowcontrol/12)
    """

    @overload
    async def __call__(self, cm: AbstractAsyncContextManager[T], /) -> T: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        /,
    ) -> tuple[T, T2]: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        cm3: AbstractAsyncContextManager[T3],
        /,
    ) -> tuple[T, T2, T3]: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        cm3: AbstractAsyncContextManager[T3],
        cm4: AbstractAsyncContextManager[T4],
        /,
    ) -> tuple[T, T2, T3, T4]: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        cm3: AbstractAsyncContextManager[T3],
        cm4: AbstractAsyncContextManager[T4],
        cm5: AbstractAsyncContextManager[T5],
        /,
    ) -> tuple[T, T2, T3, T4, T5]: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        cm3: AbstractAsyncContextManager[T3],
        cm4: AbstractAsyncContextManager[T4],
        cm5: AbstractAsyncContextManager[T5],
        cm6: AbstractAsyncContextManager[T6],
        /,
    ) -> tuple[T, T2, T3, T4, T5, T6]: ...

    async def __call__(self, *cms):
        """A quick alias for `self.enter_async_context`.

        Enter an async context manager, and schedule its exit for later.

        Accepts multiple async context managers.
        """
        if len(cms) == 1:
            return await self.enter_async_context(cms[0])
        return tuple([await self.enter_async_context(cm) for cm in cms])

    @classmethod
    def enable(
        cls, function: "Callable[Concatenate[Deferrer, P], Aw]"
    ) -> Callable[P, Aw]:
        """A decorator to be applied to a coroutine function, enabling the use of Defer.

        The coroutine function should receive an instance of `Defer` as its first
        positional argument; this will be provided by `Defer`.
        """

        async def inner(*args, **kwargs):
            defer = cls()
            async with defer:
                return await function(defer, *args, **kwargs)

        return inner


_ACTIVE_DEFER: Final[ContextVar[Union[Deferrer, None]]] = ContextVar(
    "_ACTIVE_DEFER", default=None
)


class _defer:  # noqa: N801
    """Call to defer a context manager after applying `@defer.enable` to a coroutine
    function.
    """

    @staticmethod
    def enable(function: Callable[P, Aw]) -> Callable[P, Aw]:
        """Use as a decorator on a coroutine function to enable the use of `defer`."""

        async def inner(*args, **kwargs):
            defer = Deferrer()
            token = _ACTIVE_DEFER.set(defer)
            try:
                async with defer:
                    return await function(*args, **kwargs)
            finally:
                _ACTIVE_DEFER.reset(token)

        return inner

    @overload
    async def __call__(self, cm: AbstractAsyncContextManager[T], /) -> T: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        /,
    ) -> tuple[T, T2]: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        cm3: AbstractAsyncContextManager[T3],
        /,
    ) -> tuple[T, T2, T3]: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        cm3: AbstractAsyncContextManager[T3],
        cm4: AbstractAsyncContextManager[T4],
        /,
    ) -> tuple[T, T2, T3, T4]: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        cm3: AbstractAsyncContextManager[T3],
        cm4: AbstractAsyncContextManager[T4],
        cm5: AbstractAsyncContextManager[T5],
        /,
    ) -> tuple[T, T2, T3, T4, T5]: ...

    @overload
    async def __call__(
        self,
        cm: AbstractAsyncContextManager[T],
        cm2: AbstractAsyncContextManager[T2],
        cm3: AbstractAsyncContextManager[T3],
        cm4: AbstractAsyncContextManager[T4],
        cm5: AbstractAsyncContextManager[T5],
        cm6: AbstractAsyncContextManager[T6],
        /,
    ) -> tuple[T, T2, T3, T4, T5, T6]: ...

    async def __call__(self, *args):
        active = _ACTIVE_DEFER.get()
        if active is None:
            raise Exception(
                "Defer not enabled, did you forget to apply `@defer.enable`?"
            )
        return await active(*args)



"""A better `gather`."""

from __future__ import annotations

from collections.abc import Coroutine
from typing import Any, Literal, TypeVar, overload

from ._taskgroup import TaskGroup

# Type hints taken from https://github.com/python/typeshed/blob/main/stdlib/asyncio/tasks.pyi,
# bless their hearts.
_T = TypeVar("_T")
_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")
_T5 = TypeVar("_T5")
_T6 = TypeVar("_T6")


@overload
async def gather(  # type: ignore[overload-overlap]
    coro: Coroutine[Any, Any, _T1], *, return_exceptions: Literal[False] = False
) -> tuple[_T1]: ...


@overload
async def gather(  # type: ignore[overload-overlap]
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2]: ...


@overload
async def gather(  # type: ignore[overload-overlap]
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    __coro_or_future3: Coroutine[Any, Any, _T3],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2, _T3]: ...


@overload
async def gather(  # type: ignore[overload-overlap]
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    __coro_or_future3: Coroutine[Any, Any, _T3],
    __coro_or_future4: Coroutine[Any, Any, _T4],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2, _T3, _T4]: ...


@overload
async def gather(  # type: ignore[overload-overlap]
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    __coro_or_future3: Coroutine[Any, Any, _T3],
    __coro_or_future4: Coroutine[Any, Any, _T4],
    __coro_or_future5: Coroutine[Any, Any, _T5],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2, _T3, _T4, _T5]: ...


@overload
async def gather(  # type: ignore[overload-overlap]
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    __coro_or_future3: Coroutine[Any, Any, _T3],
    __coro_or_future4: Coroutine[Any, Any, _T4],
    __coro_or_future5: Coroutine[Any, Any, _T5],
    __coro_or_future6: Coroutine[Any, Any, _T6],
    *,
    return_exceptions: Literal[False] = False,
) -> tuple[_T1, _T2, _T3, _T4, _T5, _T6]: ...


@overload
async def gather(  # type: ignore[overload-overlap]
    *coros_or_futures: Coroutine[Any, Any, _T],
    return_exceptions: Literal[False] = False,
) -> list[_T]: ...


@overload
async def gather(
    __coro_or_future1: Coroutine[Any, Any, _T1], *, return_exceptions: bool
) -> tuple[_T1 | BaseException]: ...


@overload
async def gather(
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    *,
    return_exceptions: bool,
) -> tuple[_T1 | BaseException, _T2 | BaseException]: ...


@overload
async def gather(
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    __coro_or_future3: Coroutine[Any, Any, _T3],
    *,
    return_exceptions: bool,
) -> tuple[_T1 | BaseException, _T2 | BaseException, _T3 | BaseException]: ...


@overload
async def gather(
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    __coro_or_future3: Coroutine[Any, Any, _T3],
    __coro_or_future4: Coroutine[Any, Any, _T4],
    *,
    return_exceptions: bool,
) -> tuple[
    _T1 | BaseException,
    _T2 | BaseException,
    _T3 | BaseException,
    _T4 | BaseException,
]: ...


@overload
async def gather(
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    __coro_or_future3: Coroutine[Any, Any, _T3],
    __coro_or_future4: Coroutine[Any, Any, _T4],
    __coro_or_future5: Coroutine[Any, Any, _T5],
    *,
    return_exceptions: bool,
) -> tuple[
    _T1 | BaseException,
    _T2 | BaseException,
    _T3 | BaseException,
    _T4 | BaseException,
    _T5 | BaseException,
]: ...


@overload
async def gather(
    __coro_or_future1: Coroutine[Any, Any, _T1],
    __coro_or_future2: Coroutine[Any, Any, _T2],
    __coro_or_future3: Coroutine[Any, Any, _T3],
    __coro_or_future4: Coroutine[Any, Any, _T4],
    __coro_or_future5: Coroutine[Any, Any, _T5],
    __coro_or_future6: Coroutine[Any, Any, _T6],
    *,
    return_exceptions: bool,
) -> tuple[
    _T1 | BaseException,
    _T2 | BaseException,
    _T3 | BaseException,
    _T4 | BaseException,
    _T5 | BaseException,
    _T6 | BaseException,
]: ...


@overload
async def gather(
    *coros_or_futures: Coroutine[Any, Any, _T], return_exceptions: bool
) -> list[_T | BaseException]: ...


async def gather(  # type: ignore[misc]
    *coros: Coroutine, return_exceptions: bool = False
) -> tuple:
    """A safer version of `asyncio.gather`.

    Uses a task group under the hood to not leak tasks in cases of errors
    in child tasks.

    Notable differences are:

    * If a child task fails other unfinished tasks will be cancelled, just like
      in a TaskGroup.
    * `quattro.gather` only accepts coroutines and not futures and
      generators, just like a TaskGroup.
    * When `return_exceptions` is false (the default), an exception in a child task
      will cause an ExceptionGroup to bubble out of the top-level `gather()` call,
      just like in a TaskGroup.
    * Results are returned as a tuple, not a list.

    (See https://docs.python.org/3/library/asyncio-task.html#asyncio.gather)

    .. versionadded:: 23.1.0
    """
    if not coros:
        return ()

    async with TaskGroup() as tg:
        subtasks = [
            tg.create_task(coro if not return_exceptions else _wrap_coro(coro))
            for coro in coros
        ]

    return tuple([await f for f in subtasks])


async def _wrap_coro(coro: Coroutine[Any, Any, _T]) -> _T | BaseException:
    """Adapt a raised exception into the return value."""
    try:
        return await coro
    except BaseException as exc:
        return exc



from asyncio import CancelledError, create_task, sleep
from time import monotonic, time

import pytest

from quattro import move_on_after


async def test_move_on_after():
    start = time()
    t = 0.5
    with move_on_after(t) as cancel_scope:
        await sleep(1.0)

    assert cancel_scope.cancelled_caught
    assert start + t <= time() <= start + t + 0.1


async def test_move_on_after_not_triggered():
    start = time()
    with move_on_after(0.2) as cancel_scope:
        await sleep(0.1)
    assert not cancel_scope.cancelled_caught
    assert start + 0.1 <= time() <= start + 0.2
    await sleep(0.3)  # Sleep to possibly fail due to uncancelled timers


async def test_move_on_after_not_triggered_no_await():
    with move_on_after(0.2) as cancel_scope:
        pass
    assert not cancel_scope.cancelled_caught
    await sleep(0.3)


async def test_move_on_externally_cancelled():
    cancel_scope = None

    async def task():
        nonlocal cancel_scope
        with move_on_after(0.2) as cancel_scope:
            # Will be externally cancelled here.
            await sleep(1)

    t = create_task(task())
    await sleep(0.1)
    assert not t.done()
    t.cancel()
    with pytest.raises(CancelledError):
        await t

    assert cancel_scope is not None
    assert not cancel_scope.cancelled_caught


async def test_move_on_inner_exception():
    cancel_scope = None

    async def task():
        nonlocal cancel_scope
        with move_on_after(0.3) as cancel_scope:
            await sleep(0.2)
            1 / 0  # noqa: B018

    t = create_task(task())
    await sleep(0.1)
    assert not t.done()
    with pytest.raises(ZeroDivisionError):
        await t

    assert cancel_scope is not None
    assert not cancel_scope.cancelled_caught

    await sleep(0.3)  # Catch lingering timers


async def test_move_on_nested_outer_shorter():
    """Nested move_on blocks work properly."""
    checkpt_1 = 0
    checkpt_2 = 0
    start = time()
    with move_on_after(0.1) as outer:
        checkpt_1 = 1
        with move_on_after(0.2) as inner:
            checkpt_2 = 1
            await sleep(0.3)
            pytest.fail("Not cancelled")
        pytest.fail("Not cancelled")

    assert start + 0.1 <= time() <= start + 0.15
    assert checkpt_1 == 1
    assert checkpt_2 == 1
    assert outer.cancelled_caught
    assert not inner.cancelled_caught


async def test_move_on_nested_inner_shorter():
    """Nested move_on blocks work properly."""
    checkpt_1 = 0
    checkpt_2 = 0
    checkpt_3 = 0
    start = time()
    with move_on_after(0.2) as outer:
        checkpt_1 = 1
        with move_on_after(0.1) as inner:
            checkpt_2 = 1
            await sleep(0.3)
            pytest.fail("Not cancelled")
        checkpt_3 = 1
        await sleep(0.2)

    assert start + 0.2 <= time() <= start + 0.25
    assert checkpt_1 == 1
    assert checkpt_2 == 1
    assert checkpt_3 == 1
    assert outer.cancelled_caught
    assert inner.cancelled_caught


async def test_move_on_nested_happy():
    """Nested move_on blocks work properly."""
    checkpt_1 = 0
    checkpt_2 = 0
    checkpt_3 = 0
    start = monotonic()
    with move_on_after(0.2) as outer:
        checkpt_1 = 1
        with move_on_after(0.1) as inner:
            checkpt_2 = 1
        checkpt_3 = 1

    assert start <= monotonic() <= start + 0.005
    assert checkpt_1 == 1
    assert checkpt_2 == 1
    assert checkpt_3 == 1
    assert not outer.cancelled_caught
    assert not inner.cancelled_caught


async def test_move_on_cancel_myself():
    async def task():
        with move_on_after(0.5) as cancel_scope:
            await sleep(0.1)
            cancel_scope.cancel()

    t = create_task(task())
    await sleep(0.05)
    assert not t.done()
    await sleep(0.1)
    assert t.done()

    await sleep(0.5)


async def test_move_on_precancel() -> None:
    cancel_scope = move_on_after(0.5)

    async def task():
        with cancel_scope:
            await sleep(0.1)

    cancel_scope.cancel()

    t = create_task(task())
    await sleep(0.01)
    assert t.done()

    await sleep(0.5)


async def test_move_on_after_move_deadline():
    start = time()
    with move_on_after(0.1) as cancel_scope:
        await sleep(0.01)
        cancel_scope.deadline += 0.1
        await sleep(1)

    assert start + 0.2 <= time() <= start + 0.21


async def test_move_on_after_remove_deadline():
    start = time()
    with move_on_after(0.2) as cancel_scope:
        await sleep(0.1)
        cancel_scope.deadline = None
        await sleep(0.3)

    spent = time() - start
    assert 0.4 <= spent <= 0.41


async def test_move_on_after_move_deadline_to_past():
    start = time()
    with move_on_after(0.2) as cancel_scope:
        await sleep(0.1)
        cancel_scope.deadline -= 0.1
        await sleep(0.3)

    spent = time() - start
    assert 0.1 <= spent <= 0.15


async def test_enter_twice():
    """Cannot enter twice."""
    with (c := move_on_after(0.2)), pytest.raises(RuntimeError), c:
        pass



from __future__ import annotations

import sys
from contextvars import Context
from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from asyncio import Task, _CoroutineLike
    from types import TracebackType


if sys.version_info < (3, 11):
    from taskgroup import TaskGroup as _TaskGroup
else:
    from asyncio.taskgroups import TaskGroup as _TaskGroup

__all__ = ["TaskGroup"]

T = TypeVar("T")


class TaskGroup(_TaskGroup):
    def __init__(self) -> None:
        _TaskGroup.__init__(self)
        self._bg_tasks: set[Task] = set()

    def create_background_task(
        self,
        coro: _CoroutineLike[T],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> Task[T]:
        """Create, schedule and return a task.

        When the task group shuts down, the task will be cancelled.

        If this task finishes with an error, the entire task group will be
        cancelled, like with non-background tasks.
        """
        task = self.create_task(coro, name=name, context=context)
        if not task.done():
            self._bg_tasks.add(task)
            task.add_done_callback(lambda t: self._bg_tasks.discard(t))
        return task

    async def __aexit__(
        self,
        et: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        for bg_task in self._bg_tasks:
            # Can a task be done without having executed its callbacks?
            # Yes, because callbacks get scheduled, not executed, when
            # it finishes. So a task can still be in _bg_tasks even though
            # it is `.done()`.
            if bg_task.done():
                continue

            bg_task.cancel()

        await _TaskGroup.__aexit__(self, et, exc, tb)



"""Unit tests for the background tasks."""

from __future__ import annotations

import sys
from asyncio import CancelledError, get_running_loop, sleep

import pytest
from pytest import raises

from quattro import TaskGroup

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup


async def forever():
    while True:
        await sleep(1)


async def error_out() -> None:
    await sleep(0.1)
    raise ValueError("Test")


async def return_result(after: float | None = None) -> int:
    if after is not None:
        await sleep(after)
    return 1


async def test_tasks_are_stopped():
    """Bg tasks are correctly stopped when exiting."""
    async with TaskGroup() as tg:
        task = tg.create_background_task(forever())

    assert task.done()


async def test_errors_interrupt():
    """Errors in bg tasks interrupt."""

    with raises(ExceptionGroup) as exc:
        async with TaskGroup() as tg:
            task = tg.create_background_task(forever())
            task2 = tg.create_background_task(error_out())
            await sleep(1)

    assert isinstance(exc.value.exceptions[0], ValueError)

    assert task.done()

    with raises(CancelledError):
        task.exception()

    assert task2.done()
    assert isinstance(task2.exception(), ValueError)


async def test_bg_cancellation():
    """Cancellations in bg tasks are swallowed.

    Apparently, task groups swallow CancelledErrors in their children.
    """

    async with TaskGroup() as tg:
        task = tg.create_task(forever())
        await sleep(1)
        task.cancel()
        await sleep(1)

    with raises(CancelledError):
        task.exception()


async def test_result():
    """Background tasks return proper results."""

    async with TaskGroup() as tg:
        task = tg.create_background_task(return_result())
        res = await task

    assert res == 1

    async with TaskGroup() as tg:
        task = tg.create_background_task(return_result(0.01))
        await sleep(0.02)
        res = await task

    assert res == 1

    async with TaskGroup() as tg:
        task = tg.create_background_task(forever())
        await sleep(0)

    if hasattr(task, "cancelling"):
        # Taskgroups do not clean up their cancels, and neither do we.
        assert task.cancelling() == 1

    with raises(CancelledError):
        task.exception()
    with raises(CancelledError):
        task.result()


@pytest.mark.skipif(
    sys.version_info < (3, 12),
    reason="Eager task factory only available in Python 3.12+",
)
async def test_eager_task_factory() -> None:
    """Eager task factories behave properly."""
    from asyncio import eager_task_factory

    get_running_loop().set_task_factory(eager_task_factory)

    async with TaskGroup() as tg:
        tg.create_background_task(return_result())



import sys
from asyncio import CancelledError, current_task, get_running_loop, sleep
from asyncio import gather as asyncio_gather

from pytest import mark, raises

from quattro import gather

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup


@mark.parametrize("gather", [gather, asyncio_gather])
async def test_empty(gather):
    """An empty gather works."""
    # asyncio gather returns a list
    assert tuple(await gather()) == ()


@mark.parametrize("gather", [gather, asyncio_gather])
async def test_simple_gather(gather):
    """Simple gather works."""

    async def test() -> int:
        await sleep(0.01)
        return 1

    assert tuple(await gather(test(), test())) == (1, 1)


@mark.parametrize("gather", [gather, asyncio_gather])
async def test_gather_with_error(gather):
    """Gather works if there's an error."""
    cancelled = 0

    async def test() -> int:
        nonlocal cancelled
        try:
            await sleep(0.1)
        except CancelledError:
            cancelled += 1
        return 1

    async def error() -> None:
        await sleep(0.005)
        raise ValueError()

    with raises((ExceptionGroup, ValueError)) as exc_info:
        await gather(test(), test(), error())

    if gather == asyncio_gather:
        assert isinstance(exc_info.value, ValueError)
        # default asyncio behavior
        assert cancelled == 0
    else:
        assert isinstance(exc_info.value, ExceptionGroup)
        assert repr(exc_info.value.exceptions[0]) == "ValueError()"
        assert cancelled == 2


@mark.parametrize("gather", [gather, asyncio_gather])
async def test_simple_gather_exceptions(gather):
    """Simple gather works when collecting exceptions."""

    async def test() -> int:
        await sleep(0.01)
        return 1

    assert tuple(await gather(test(), test(), return_exceptions=True)) == (1, 1)


@mark.parametrize("gather", [gather, asyncio_gather])
async def test_with_error_return_excs(gather):
    """Gather works if there's an error and exceptions are returned."""

    async def test() -> int:
        await sleep(0.01)
        return 1

    err = ValueError()

    async def error() -> None:
        await sleep(0.005)
        raise err

    res = await gather(test(), test(), error(), return_exceptions=True)

    assert tuple(res) == (1, 1, err)


@mark.parametrize("gather", [gather, asyncio_gather])
async def test_parent_cancelled(gather):
    """When the parent is cancelled, the children are also cancelled."""
    cancelled = 0

    async def test() -> int:
        nonlocal cancelled
        try:
            await sleep(0.01)
        except CancelledError:
            cancelled += 1
            raise
        return 1

    res = None

    # We cannot use `move_on` here since asyncio.gather doesn't
    # work with it on some versions of 3.9 and 3.10.
    current = current_task()
    get_running_loop().call_later(0.001, lambda: current.cancel())

    with raises(CancelledError):
        res = await gather(test(), test())

    assert res is None
    assert cancelled == 2


@mark.parametrize("gather", [gather, asyncio_gather])
async def test_parent_cancelled_return_excs(gather):
    """When the parent is cancelled, the children are also cancelled."""
    cancelled = 0

    async def test() -> int:
        nonlocal cancelled
        try:
            await sleep(0.01)
        except CancelledError:
            cancelled += 1
            raise
        return 1

    res = None

    # We cannot use
    # `move_on` here since asyncio.gather doesn't
    # work with it on some versions of 3.9 and 3.10.
    current = current_task()
    get_running_loop().call_later(0.001, lambda: current.cancel())

    with raises(CancelledError):
        res = await gather(test(), test(), return_exceptions=True)

    assert res is None
    assert cancelled == 2



from contextlib import asynccontextmanager, contextmanager

from pytest import raises

from quattro import Deferrer, defer


async def test_defer_async() -> None:
    """Async defer works."""

    entered = False
    exited = False

    @asynccontextmanager
    async def asynccm():
        nonlocal entered, exited
        entered = True
        yield 1
        exited = True

    @Deferrer.enable
    async def coro(defer: Deferrer, a: int) -> int:
        assert not entered
        assert await defer(asynccm()) == 1
        assert entered
        return a

    assert await coro(1) == 1

    assert exited


async def test_defer_async_several() -> None:
    """Async defer works with multiple contexts."""

    entered = [False, False]
    exited = [False, False]

    @asynccontextmanager
    async def asynccm():
        nonlocal entered, exited
        entered[0] = True
        yield 1
        exited[0] = True

    @asynccontextmanager
    async def asynccm2():
        nonlocal entered, exited
        entered[1] = True
        yield 2
        exited[1] = True

    @Deferrer.enable
    async def coro(defer: Deferrer) -> None:
        assert not entered[0]
        assert not entered[1]
        assert await defer(asynccm(), asynccm2()) == (1, 2)
        assert entered[0]
        assert entered[1]

    await coro()

    assert exited[0]
    assert exited[1]


async def test_defer_sync_in_async() -> None:
    """Sync defer in coroutines works."""

    entered = False
    exited = False

    @contextmanager
    def cm():
        nonlocal entered, exited
        entered = True
        yield 1
        exited = True

    @Deferrer.enable
    async def coro(defer: Deferrer) -> None:
        assert not entered
        assert defer.enter_context(cm())
        assert entered

    await coro()

    assert exited


async def test_defer() -> None:
    """The `defer` decorator works."""

    entered = False
    exited = False

    @asynccontextmanager
    async def asynccm():
        nonlocal entered, exited
        entered = True
        yield 1
        exited = True

    @defer.enable
    async def coro(a: int) -> int:
        assert not entered
        assert await defer(asynccm()) == 1
        assert entered
        return a

    assert await coro(1) == 1

    assert exited


async def test_defer_nested() -> None:
    """The `defer` decorator works on nested coroutines."""

    entered = False
    entered2 = False
    exited = False
    exited2 = False

    @asynccontextmanager
    async def asynccm():
        nonlocal entered, exited
        entered = True
        yield 1
        exited = True

    @asynccontextmanager
    async def asynccm2():
        nonlocal entered2, exited2
        entered2 = True
        yield 2
        exited2 = True

    @defer.enable
    async def coro(a: int) -> int:
        assert not entered
        assert await defer(asynccm()) == 1
        assert entered
        return a

    @defer.enable
    async def coro2(b: int) -> int:
        assert not entered
        assert not entered2
        assert await defer(asynccm2()) == 2
        assert not entered
        assert entered2

        assert not exited
        assert not exited2

        await coro(b)

        assert exited
        assert not exited2
        return b

    assert await coro2(1) == 1

    assert exited
    assert exited2


async def test_defer_no_decorator() -> None:
    """Forgetting the decorator raises."""

    entered = False
    exited = False

    @asynccontextmanager
    async def asynccm():
        nonlocal entered, exited
        entered = True
        yield 1
        exited = True

    async def coro(a: int) -> int:
        assert not entered
        with raises(Exception) as exc:
            assert await defer(asynccm()) == 1
        raise exc.value

    with raises(Exception) as exc:
        await coro(1)

    assert "Defer not enabled" in exc.value.args[0]



"""Task control for asyncio."""

from __future__ import annotations

from typing import Final

from ._cancelscope import (
    CancelScope,
    cancel_stack,
    fail_after,
    fail_at,
    move_on_after,
    move_on_at,
)
from ._defer import Deferrer, _defer
from ._gather import gather
from ._taskgroup import TaskGroup

__all__ = [
    "CancelScope",
    "Deferrer",
    "TaskGroup",
    "defer",
    "fail_after",
    "fail_at",
    "gather",
    "get_current_effective_deadline",
    "move_on_after",
    "move_on_at",
]


def get_current_effective_deadline() -> float:
    return min(
        [cs._deadline for cs in cancel_stack.get() if cs._deadline is not None],
        default=float("inf"),
    )


# This needs to be here for Sphinx.
defer: Final = _defer()
"""First wrap your coroutine function with `defer.enable`, then call me inside."""
